#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016.

# Author(s):

#   Christian Kliche <christian.kliche@ebp.de>

# This file is part of the mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.
'''
Module for writer option constants
'''
WR_OPT_NBITS = 'nbits'
WR_OPT_COMPRESSION = 'compression'
WR_OPT_BLOCKSIZE = 'blocksize'
WR_OPT_FILL_VALUE_SUBST = 'fill_value_subst'
